﻿// ==================================================================
// Function TUTORIAL for Demonstrating the Matrix Classes
//          --------
// C++ - MATRIX OAjects and Methods for Tensor/Matrix Calculus
// ==================================================================
// Copyright: Univ.-Prof. Dr.-Ing. Udo F. Meissner, Erkrath/Darmstadt
// Advanced teaching and study program - Commercial use not permitted.
// The author assumes no liaAility for incompleteness and errors!
// Version:  2023
// Compiler: C++ 
// ==================================================================
// TUTORIAL: Generation of Matrix Objects and
//           Handling of elementary Over-Loaded Operators 
// =============== Include Files ====================================
#include "pch.h"                          // pre-compiled header file
// -------------
// #include "Matrizen.h" 
            // aAstract Matrices class                   MATRIZEN
// #include "Matrix.h" 
            // 4-dimensional douAle rectangle matrices   MATRIX
// #include "Vektor.h" 
            // 1-dimensional vectors                     VEKTOR
// #include "delta_Matrix.h"  
            // delta-matrices                            delta_MATRIX
// #include "e_Matrix.h"   
            // e-matrices                                e_MATRIX
// #include "intMatrix.h" 
            // 4-dimensional integer matrices            intMATRIX
// #include "intVektor.h"  
            // 1-dimensional integer vectors             intVEKTOR
// ==================================================================
//  TUTORIUM 
int TUTORIAL(ofstream& outFILE, ifstream& inFILE)
{                         // outFILE logs the data input 
                          // inFile provides the data input
    // --------------------------------------------------------------
    // ofstream outFILE;
    // string output = "Output file";
    // outFILE.open(output, ios::app);
    // std::cout << "Output file: " << Output << endl << endl;
    // ==============================================================
    bool ERROR;
    ERROR = false;
    // ERROR = true;
    if (ERROR) std::cout << "Error = " << ERROR << endl;
    // ======================== TUTORIAL ==============================
    cout << endl
         << "1.1 - Matrix Objects and elementary Operators" << endl
         << "=============================================" << endl << endl;
    //
    // Preallocate of C-array with values
    // ----------------------------------
    double F[10] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 };
    std::cout << "C-Vector F(10) = ";
    for (int j = 0; j < 10; j++) std::cout << " " << F[j];           // printing
    std::cout << endl << endl;

    // Creating a static instance of the class VEKTOR using constructor
    // ----------------------------------------------------------------
    VEKTOR V(F, 6);                       // (1x6)-vector V is preloaded with F
    std::cout << "(1x6) Vector V := F = " << V << endl;
                                                     // overloaded operator '<<'
    // Creating a static instance of the class MATRIX using constructor
    // ----------------------------------------------------------------
    MATRIX M(3, 3); // (3x3)-matrix, assignment of V to M
    // Exception: type casting by M := V for demonstration
    M = V;                                            // overloaded operator '='
    std::cout << "(3x3) Matrix M := V = " << M << endl;              // printing

    // Creating a static instance of class delta_MATRIX using constructor
    // ------------------------------------------------------------------
    delta_MATRIX d(2, 2);                             // (2x2)-delta-matrix
    std::cout << "(2x2) delta-Matrix d = " << d << endl;             // printing

    // Creating a static instance of the MATRIX class using constructor
    // ----------------------------------------------------------------
    MATRIX A(2, 2); // (2x2)-matrix
    std::cout << "(2x2) Matrix A = " << A << endl;                   // printing
    // ================================================================
    cout << endl
         << "1.2 - Addition and Subtraction" << endl
         << "==============================" << endl << endl;
    //
    // Matrix M is temporarily resized by the member function dimension!
    // --------------------------
    // Matrix Addition M(i,k) + A(i,k)
    // ---------------------------------
    A = M.dimension(2, 2) + d;                   // overloaded operator '+'
    std::cout << "Matrix A = A + d = " << A << endl;                    // print
    std::cout << "Original (3x3) Matrix preserved: M = " << M << endl;  // print
    //
    // Matrix Subtraction A(i,k) - d(i,k)
    // -----------------------------------
    A = A - d;                                   // overloaded operator '-'
    std::cout << "Matrix A = A - d = " << A << endl;                    // print
    //
    // Determinant of Matrix A
    // -----------------------
    double S;
    // S = A.determinant (2);                                        // upcoming
       S = A.determinante(2);
    std::cout << "Determinant det(A) = " << S << endl;
    // ================================================================
    cout << endl
    << "2.1 - Cartesian Basis and Matrix/Vector Multiplications" << endl
    << "=======================================================" << endl <<endl;
    //
    // => Vectors v(i) are internally treated as matrices v(0,i)
    //    with N(0) = 1, in order to build scalar products rather easy !
    // -----------------------------------------------------------------
    // Creating three Cartesian reference vectors  
    VEKTOR e0(3), e1(3), e2(3);                                // (1x3)-Vektoren
    e0(0) = 1.0;  e1(1) = 1.0;
    std::cout << "(1x3) Vektor e0 = " << e0 << endl;                    // print
    std::cout << "(1x3) Vektor e1 = " << e1 << endl;                    // print
    //
    // Creating two physical vectors
    VEKTOR v0(3), v1(3);                                        // (1x3)-Vectors
    v0(0) = 6.0; v0(1) = 2.0;
    std::cout << "(1x3) Vektor v0 = " << v0 << endl;                    // print
    v1(0) = 1.0; v1(1) = 2.0;
    std::cout << "(1x3) Vektor v1 = " << v1 << endl;                    // print
    //
    // Calculating the scalar product s = v0(0,s) * v1(0,s)
    // ---------------------------------------------------- 
    double s;
    s = v0 * v1;
    std::cout << "Scalar s = v0 * v1 = " << s << endl << endl;          // print
    //
    // Calculating the dyadic product  M(i,k) = v0(i) * v1(k)
    // ------------------------------------------------------ 
    M = v0.dimension(3) * v1.dimension(3);
    std::cout << "Dyadic product M = v0 * v1 = " << M << endl;          // print
    //
    // Calculating the vector product
    //                             e2(0,i) = e(ikl) * e0(0,k) * e1(0,l)
    // ----------------------------------------------------------------
    // Creating a static instance of the class e_MATRIX
    //                                            using the constructor
    e_MATRIX e(3, 3, 3);                                     // (3x3x3)-e-Matrix
    std::cout << "(3x3x3) e-Matrix e = " << e << endl;                // print
    // Chained multiplication without member function dimension
    e2 = (e * e0) * e1;
    std::cout << "(1x3) Vector e2 = e0 x e1 = " << e2 << endl;          // print
    //
    // General matrix product C(i,k,l,m) = A(i,s,k) * B(l,s,m)
    // -------------------------------------------------------
    //    Using the member function M.dimension(N{i}, N{k}, N{l}, N{m}),
    // => Matrix M(s,k) can be temporarily extended by a 0-Index
    //    M(s,k) =: M(0,s,k) mit N{0} = 1,
    // => Matrix indices M(i,k,l) can be bundled up to M(IK,s,l)
    //    M(i,k,s,l) =: M(IK,s,l)  mit N{IK} = N{i} * N{k},
    // => in order to bring the index s in the centered position M(i,s,l) !   
    // --------------------------------------------------------------------
    std::cout << "(2x2) Matrix A = " << A << endl;                      // print
    //
    // Creating the matrices C and B
    MATRIX C(2, 2), B(2, 2);                                   // (2x2) Matrices
    B = d;   B(0, 1) = 0.5;
    std::cout << "(2x2) Matrix B = " << B << endl;                      // print
    // 
    // Matrix multiplication C(i,k) = A(i,s) * B(0,s,k)
    // ------------------------------------------------
    C = A * B.dimension(1, 2, 2);
    std::cout << "(2x2) Matrix C = A * B  = " << C << endl;             // print
    // 
    // Matrix multiplication C(i,k) = A(0,s,i) * B(0,s,k)
    // --------------------------------------------------
    C = A.dimension(1, 2, 2) * B.dimension(1, 2, 2);
    std::cout << "(2x2) Matrix C = A(T) * B  = " << C << endl;          // print
    //
    // Grassmann Identity
    // ------------------
    // delta-Matrix(i,k,l,m) = e-Matrix(i,k,s) * e-Matrix(l,m,s)
    // Creating a static instance of the class MATRIX
    //                                            using the constructor
    MATRIX D4(3, 3, 3, 3);                 // (3x3x3x3) Matrix         
    D4 = e.dimension(3 * 3, 3) * e.dimension(3 * 3, 3);
    std::cout << "(3x3x3x3) delta-Matrix of 4. order D4(i,k,l,m) = "
        << D4 << endl;                                            // print
    //
    // Test
    // ----
    // Creating a dynamic instance of the class delta_MATRIX
    //                                            using the constructor
    //*** delta_MATRIX& d4 = *new delta_MATRIX(3, 3, 3, 3);
                                                       // (3x3x3x3) delta-Matrix 
    //*** std::cout<<"Test: (3x3x3x3) delta-Matrix of 4. order d4(i,k,l,m) =
    //***         " << d4 << endl; // print
    //
    // Release the dynamic matrix d4
    //*** delete & d4;                            // delete using destructor '~'
    //
    // ================================================================
    cout << endl
         << "3.1 - Basis Vectors and Metrics in Euclidian Space" << endl
         << "==================================================" << endl << endl;
    // 
    // Covariant basis vectors
    VEKTOR g0(3), g1(3);
    g0 = v0;                                                   //.dimension(3);
    g1 = v1;                                                   //.dimension(3);
    //
    // Create a dynamic instance of the class BASIS using constructor
    // --------------------------------------------------------------
    //
    BASIS& basis = *new BASIS(g0, g1);
    //
    // Covariant basis vectors
    // ------------------------
    MATRIX cov_basis(3, 2);                                      // (3x2)-Matrix
    //  cov_basis = covariante_basis();                              // upcoming
        cov_basis = basis.kovariante_basis();
    std::cout << "Covariant basis vectors: (3x2) Matrix cov_basis = "
        << cov_basis << endl;                                          // print
    //
    // Covariant metric
    // -----------------
    MATRIX cov_metrik(2, 2);                                    // (2x2) Matrix 
    //  double g = basis.covariante_metrik(cov_metrik);             // upcoming
        double g = basis.kovariante_metrik(cov_metrik);
    std::cout << "Covariant metric: (2x2) Matrix cov_metrik = "
        << cov_metrik << endl;  // print
    std::cout << "Covariant metric: Determinant g = " << g << endl << endl;
    // print
    //
    // Contravariant metric
    // ---------------------
    MATRIX contra_metrik(2, 2);                                 // (2x2) Matrix
    //  basis.contravariante_metrik(contra_metrik);                 // upcoming
        basis.kontravariante_metrik(contra_metrik);
    std::cout << "Contravariant metric: (2x2) Matrix contra_metrik "
        << contra_metrik << endl;                                      // print
    //
    // Test -> Identity matrix
    // -----------------------
    MATRIX& Einh = *new MATRIX(2, 2);
    Einh = cov_metrik.dimension(1, 2, 2) * contra_metrik.dimension(1, 2, 2);
    std::cout << "Test: Identity matrix = cov_metrik * contra_metrik "
        << Einh << endl;                                               // print
    //
    // Contravariant basis vectors
    // ----------------------------
    MATRIX contra_basis(3, 2);                                  // (3x2) Matrix
    //  contra_basis = basis.contravariante_basis();                // upcoming
        contra_basis = basis.kontravariante_basis();
    std::cout << "Contravariant basis vectors: (3x2) Matrix contra_basis = "
        << contra_basis << endl;                                       // print
    //
    // Test -> Identity matrix
    // ------------------------
    Einh = *new MATRIX(2, 2);         // dynamic (2x2) Matrix
    Einh = cov_basis.dimension(1, 3, 2) * contra_basis.dimension(1, 3, 2);
    std::cout << "Test: Identity matrix = cov_basis * contra_basis "
        << Einh << endl;                                              // print
    // 
    // Release dynamic objects
    // -----------------------
    delete& basis;                           // delete using destructor '~'
    delete& Einh;                            // delete using destructor '~'
    // ================================================================
    cout << endl
         << "4.1 - Tensor Transformations" << endl
         << "============================" << endl << endl;
    // 
    VEKTOR v(3);                                                 // (1x3)-Vektor
    // Vector in cartesian reference system e0-e1-e2
    v(0) = 0.0;    v(1) = 5.0;
    cout << "Cartesian system: Vector v(i) = " << v << endl;            // print
    //
    // Transform vector v(i) into skew-angled reference system g0-g1    
    // -------------------------------------------------------------
    VEKTOR v_cov(2);                                             // (1x2) vector
    // v_cov(0,alpha) = cov_basis(0,i,alpha) * v(0,i)
    //
    v_cov = cov_basis.dimension(1, 3, 2) * v;
    cout << "Skew-angled system: Covariant vector v(alpha) = "
        << v_cov << endl;                                               // print
    //
    //Transform vector v_schief(alpha) from skew-angled to Cartesian system
    // --------------------------------------------------------------------
    // v(0,i) = contra_basis(i,alpha) * v_cov(0,alpha)  
    //
    v = contra_basis * v_cov;
    cout << "Test: Cartesian system: Vector v(i) = " << v << endl;      // print
    //
    // Transform 4th-order tensor
    // --------------------------
    // Elasticity tensor E in Cartesian coordinate system e0-e1
    // Allocated C-field with fictitious values row by row (!)
    // --------------------------------------------------------
    double F16[16] = { 1,0,0,0, 0,0.5,0.5,0, 0, 0.5, 0.5, 0, 0,0,0,1 };
    // Create a static instance of the MATRIX class using constructor
    // ----------------------------------------------------------------
    MATRIX E(F16, 2, 2, 2, 2);      // (2x2x2x2) matrix pre-filled with F16
    cout <<"(2x2x2x2) Elasticity tensor E := F16 = " << E << endl; // print
    MATRIX E_contra(2, 2, 2, 2);                             // (2x2x2x2)-Matrix
    // 
    // Transform tensor E(i,k,l,m) into skew-angled reference system g0-g1
    // -------------------------------------------------------------------
    // Reduce contravariant basis to (2x2) matrix - eliminate third row
    // contra_basis = contra_basis.submatrix_out(0,2,0,2);           // upcoming 
       contra_basis = contra_basis.untermatrix_heraus(0,2,0,2);
    cout << "Contravariant basis vectors: (2x2) Matrix contra_basis = "
        << contra_basis.dimension(2, 2) << endl;                        // print
    //
    // E_contra(i,k, gamma,delta) = (E(IK,l,m) * contra_basis(1,l,gamma)) *
    //               contra_basis(0, m, delta)    // with N{IK} = N{i}*N{k}
    //
    E_contra = E.dimension(2 * 2, 2, 2) *
                                    contra_basis.dimension(1, 2, 2);
    E_contra = E_contra.dimension(2 * 2, 2, 2) *
                                    contra_basis.dimension(1, 2, 2);
    //    
    // E_contra(alpha,beta, gamma,delta) =
    //  contra_basis(0,i,alpha) * contra_basis(0,k,beta) * E_contra(i,k,GD)
    //                                       with N{GD} = N{gamma}*N{delta}
    //
    E_contra = contra_basis.dimension(1, 2, 2) *
        E_contra.dimension(2, 2, 2 * 2);
    E_contra = contra_basis.dimension(1, 2, 2) *
        E_contra.dimension(2, 2, 2 * 2);
    cout << "Skew-angled system: Tensor E_contra(alpha,beta,gamma,delta) = "
        << E_contra << endl;                                            // print
    //
    // Generate contravariant (2x2) Levi-Civita tensor using class e_MATRIX
    // --------------------------------------------------------------------
    e_MATRIX epsilon_Tensor(2, 2, sqrt(g));
    cout << "Skew-angled system: Contravariant epsilon-Tensor = "
        << epsilon_Tensor << endl;                                     // print
    // =======================================================
    std::cout << "End Tutorial" << endl;
    std::cout << "============" << endl;
    //
    std::cout << "Pause:" << endl;
    // ###########################
    char c;
    std::cin >> c;                              outFILE << c << endl;
    // ==============================================================
    return 0;
}
